""" git-copy.py

Synopsis
--------

This script implements a two-way copy of files in git using
`git mv` to preserve history for `git blame`.

$ python git-copy.py FILES

Details
-------

For a given file -- say `myfile.py` -- we create three copies in three branches
using the branch `copied` as a base and target (which will also be created).

Branch `new_files` contains `new_myfile.py`
Branch `old_files` contains `old_myfile.py`
Branch `preserved_files` contains `preserved_myfile.py`

Then we use a git octo-merge to merge these three branches into the `copied`
branch, this results in three "copies" (two copies and one original).

Lastly, we remove the prefix `preserved_` again to restore the original
`myfile.py` in a final commit. This is a 'trick' used to restore and hence
retain the original `myfile.py`. And then the temporary branches are deleted
such that only the base branch `copied` remains.

If the script succeeds, the following files will be present:

new_myfile.py
old_myfile.py
myfile.py

.. and all three will have sane `git blame` output.

The script is rudimentary and doesn't fail gracefully or clean up in case it
fails. If something goes wrong, read the output of the error message closely.

Tip: A good way to undo the created branches if something does go wrong is to
use:

$ git checkout main
$ git branch -D copied old_files, new_files, preserved_files

... assuming `main` is your main development branch.

Lastly, You can hand over as many files as your shell and Python implementation
will handle. Hopefully this should suffice for all practical purposes.

Dependencies
------------

* https://pypi.org/project/sh/

See also
--------

This script was inspired by a technique mentioned in:

https://stackoverflow.com/questions/16937359/git-copy-file-preserving-history/44036771#44036771

"""

import os
import sys

import sh

BASE_BRANCH = "copied"
FILES = sys.argv[1:]

NEW_BRANCH_NAME = "new_files"
NEW_PREFIX = "new_"
OLD_BRANCH_NAME = "old_files"
OLD_PREFIX = "old_"
PRESERVED_BRANCH_NAME = "preserved_files"
PRESERVED_PREFIX = "preserved_"


def copy_files(branch: str, prefix: str, remove_prefix: bool = False) -> None:
    """ copy files on a branch using git mv

    This function will create a git branch and then copy all files in global
    FILES to insert or remove a prefix into or from the filename. Lastly, those
    changes will be committed to git.

    branch: the name of the branch
    prefix: the prefix to insert or remove
    remove_prefix: insert or remove

    """
    sh.git.checkout(BASE_BRANCH)
    if not remove_prefix:
        print(f"Setting up git branch: '{branch}'")
        sh.git.branch(branch)
        sh.git.checkout(branch)
    for file in FILES:
        head, tail = os.path.split(file)
        if remove_prefix:
            origin = os.path.join(head, f"{prefix}{tail}")
            print(f"git mv {origin} {file}")
            sh.git.mv(origin, file)
        else:
            move_target = os.path.join(head, f"{prefix}{tail}")
            print(f"git mv {file} {move_target}")
            sh.git.mv(file, move_target)
    if remove_prefix:
        print(f"Creating commit to remove prefix: '{prefix}'")
        sh.git.commit("--no-verify", "-m", f"remove prefix: '{prefix}'")
    else:
        print(f"Creating commit for prefix: '{prefix}'")
        sh.git.commit("--no-verify", "-m", f"copy files to prefix: '{prefix}'")


if __name__ == "__main__":

    # exit if no arguments given
    if len(FILES) == 0:
        print("usage: git-copy FILES")
        sys.exit(1)

    # setting up the `copied` branch
    print(f"Setting up git branch: '{BASE_BRANCH}'")
    sh.git.branch(BASE_BRANCH)
    sh.git.checkout(BASE_BRANCH)

    # create the three branches and do the git mv and commit
    copy_files(NEW_BRANCH_NAME, NEW_PREFIX)
    copy_files(OLD_BRANCH_NAME, OLD_PREFIX)
    copy_files(PRESERVED_BRANCH_NAME, PRESERVED_PREFIX)

    # perform the octo-merge of the three branches
    sh.git.checkout(BASE_BRANCH)
    sh.git.merge(NEW_BRANCH_NAME, OLD_BRANCH_NAME, PRESERVED_BRANCH_NAME)

    # restore the originals
    copy_files(PRESERVED_BRANCH_NAME, PRESERVED_PREFIX, remove_prefix=True)

    # remove stale branches
    sh.git.branch("-d", NEW_BRANCH_NAME, OLD_BRANCH_NAME, PRESERVED_BRANCH_NAME)

    print("...done")



#! /usr/bin/env python

import sys
import os

from github3 import login
import github3


def fetch(orgname, reponame, last_num, gh):

    repo = gh.repository(orgname, reponame)
    issues = repo.issues(state='all')

    opened_issues = []
    closed_issues = []

    opened_prs = []
    closed_prs = []

    max_iss_num = 0
    for issue in issues:
        info = issue.as_dict()
        iss_num = int(info['number'])
        max_iss_num = max(max_iss_num, iss_num)

        if iss_num <= last_num:
            break
        merged = False
        if issue.pull_request_urls:
            # Is PR?
            merged = bool(info['pull_request'].get("merged_at"))
            where = {'opened': opened_prs, 'closed': closed_prs}
        else:
            # Is Issues
            where = {'opened': opened_issues, 'closed': closed_issues}

        line = f"{' - merged ' if merged else ''}- [{reponame}"\
               f"#{info['number']}]({info['html_url']}) - {info['title']}"

        # Is issue already merged
        if issue.is_closed():
            where['closed'].append(line)
        else:
            where['opened'].append(line)

    return {
        'opened_issues': opened_issues,
        'closed_issues': closed_issues,
        'opened_prs': opened_prs,
        'closed_prs': closed_prs,
        'max_iss_num': max_iss_num,
    }


def display(data):
    print("## 1. New Issues")
    for line in reversed(data['opened_issues']):
        print(line)
    print()

    print("### Closed Issues")
    for line in reversed(data['closed_issues']):
        print(line)
    print()

    print("## 2. New PRs")
    for line in reversed(data['opened_prs']):
        print(line)
    print()

    print("### Closed PRs")
    for line in reversed(data['closed_prs']):
        print(line)
    print()


def main(numba_last_num, llvmlite_last_num, user=None, password=None):

    if user is not None and password is not None:
        gh = login(str(user), password=str(password))
    else:
        gh = github3

    numba_data = fetch("numba", "numba", numba_last_num, gh)
    llvmlite_data = fetch("numba", "llvmlite", llvmlite_last_num, gh)

    # combine data
    data = {
        'opened_issues':
            llvmlite_data['opened_issues'] +
            numba_data['opened_issues'],
        'closed_issues':
            llvmlite_data['closed_issues'] +
            numba_data['closed_issues'],
        'opened_prs':
            llvmlite_data['opened_prs'] +
            numba_data['opened_prs'],
        'closed_prs':
            llvmlite_data['closed_prs'] +
        numba_data['closed_prs'],
    }

    display(data)

    print(f"(last numba: {numba_data['max_iss_num']};"
          f"llvmlite {llvmlite_data['max_iss_num']})")


help_msg = """
Usage:
    {program_name} <numba_last_num> <llvmlite_last_num>
"""

if __name__ == '__main__':
    program_name = sys.argv[0]
    try:
        [numba_last_num, llvmlite_last_num] = sys.argv[1:]
    except ValueError:
        print(help_msg.format(program_name=program_name))
    else:
        main(int(numba_last_num),
             int(llvmlite_last_num),
             user=os.environ.get("GHUSER"),
             password=os.environ.get("GHPASS"))



import numpy as np
import operator
from collections import namedtuple

from numba.core import types, utils
from numba.core.typing.templates import (AttributeTemplate, AbstractTemplate,
                                         infer, infer_global, infer_getattr,
                                         signature, bound_function)
# import time side effect: array operations requires typing support of sequence
# defined in collections: e.g. array.shape[i]
from numba.core.typing import collections
from numba.core.errors import (TypingError, RequireLiteralValue, NumbaTypeError,
                               NumbaNotImplementedError, NumbaAssertionError,
                               NumbaKeyError, NumbaIndexError, NumbaValueError)
from numba.core.cgutils import is_nonelike

numpy_version = tuple(map(int, np.__version__.split('.')[:2]))


Indexing = namedtuple("Indexing", ("index", "result", "advanced"))


def get_array_index_type(ary, idx):
    """
    Returns None or a tuple-3 for the types of the input array, index, and
    resulting type of ``array[index]``.

    Note: This is shared logic for ndarray getitem and setitem.
    """
    if not isinstance(ary, types.Buffer):
        return

    ndim = ary.ndim

    left_indices = []
    right_indices = []
    ellipsis_met = False
    advanced = False
    num_newaxis = 0

    if not isinstance(idx, types.BaseTuple):
        idx = [idx]

    # Here, a subspace is considered as a contiguous group of advanced indices.
    # num_subspaces keeps track of the number of such
    # contiguous groups.
    in_subspace = False
    num_subspaces = 0
    array_indices = 0

    # Walk indices
    for ty in idx:
        if ty is types.ellipsis:
            if ellipsis_met:
                raise NumbaTypeError(
                    "Only one ellipsis allowed in array indices "
                    "(got %s)" % (idx,))
            ellipsis_met = True
            in_subspace = False
        elif isinstance(ty, types.SliceType):
            # If we encounter a non-advanced index while in a
            # subspace then that subspace ends.
            in_subspace = False
        # In advanced indexing, any index broadcastable to an
        # array is considered an advanced index. Hence all the
        # branches below are considered as advanced indices.
        elif isinstance(ty, types.Integer):
            # Normalize integer index
            ty = types.intp if ty.signed else types.uintp
            # Integer indexing removes the given dimension
            ndim -= 1
            # If we're within a subspace/contiguous group of
            # advanced indices then no action is necessary
            # since we've already counted that subspace once.
            if not in_subspace:
                # If we're not within a subspace and we encounter
                # this branch then we have a new subspace/group.
                num_subspaces += 1
                in_subspace = True
        elif (isinstance(ty, types.Array) and ty.ndim == 0
              and isinstance(ty.dtype, types.Integer)):
            # 0-d array used as integer index
            ndim -= 1
            if not in_subspace:
                num_subspaces += 1
                in_subspace = True
        elif (isinstance(ty, types.Array)
              and isinstance(ty.dtype, (types.Integer, types.Boolean))):
            if ty.ndim > 1:
                # Advanced indexing limitation # 1
                raise NumbaTypeError(
                    "Multi-dimensional indices are not supported.")
            array_indices += 1
            # The condition for activating advanced indexing is simply
            # having at least one array with size > 1.
            advanced = True
            if not in_subspace:
                num_subspaces += 1
                in_subspace = True
        elif (is_nonelike(ty)):
            ndim += 1
            num_newaxis += 1
        else:
            raise NumbaTypeError("Unsupported array index type %s in %s"
                                 % (ty, idx))
        (right_indices if ellipsis_met else left_indices).append(ty)

    if advanced:
        if array_indices > 1:
            # Advanced indexing limitation # 2
            msg = "Using more than one non-scalar array index is unsupported."
            raise NumbaTypeError(msg)

        if num_subspaces > 1:
            # Advanced indexing limitation # 3
            msg = ("Using more than one indexing subspace is unsupported."
                   " An indexing subspace is a group of one or more"
                   " consecutive indices comprising integer or array types.")
            raise NumbaTypeError(msg)

    # Only Numpy arrays support advanced indexing
    if advanced and not isinstance(ary, types.Array):
        return

    # Check indices and result dimensionality
    all_indices = left_indices + right_indices
    if ellipsis_met:
        assert right_indices[0] is types.ellipsis
        del right_indices[0]

    n_indices = len(all_indices) - ellipsis_met - num_newaxis
    if n_indices > ary.ndim:
        raise NumbaTypeError("cannot index %s with %d indices: %s"
                             % (ary, n_indices, idx))
    if n_indices == ary.ndim and ndim == 0 and not ellipsis_met:
        # Full integer indexing => scalar result
        # (note if ellipsis is present, a 0-d view is returned instead)
        res = ary.dtype

    elif advanced:
        # Result is a copy
        res = ary.copy(ndim=ndim, layout='C', readonly=False)

    else:
        # Result is a view
        if ary.slice_is_copy:
            # Avoid view semantics when the original type creates a copy
            # when slicing.
            return

        # Infer layout
        layout = ary.layout

        def keeps_contiguity(ty, is_innermost):
            # A slice can only keep an array contiguous if it is the
            # innermost index and it is not strided
            return (ty is types.ellipsis or isinstance(ty, types.Integer)
                    or (is_innermost and isinstance(ty, types.SliceType)
                        and not ty.has_step))

        def check_contiguity(outer_indices):
            """
            Whether indexing with the given indices (from outer to inner in
            physical layout order) can keep an array contiguous.
            """
            for ty in outer_indices[:-1]:
                if not keeps_contiguity(ty, False):
                    return False
            if outer_indices and not keeps_contiguity(outer_indices[-1], True):
                return False
            return True

        if layout == 'C':
            # Integer indexing on the left keeps the array C-contiguous
            if n_indices == ary.ndim:
                # If all indices are there, ellipsis's place is indifferent
                left_indices = left_indices + right_indices
                right_indices = []
            if right_indices:
                layout = 'A'
            elif not check_contiguity(left_indices):
                layout = 'A'
        elif layout == 'F':
            # Integer indexing on the right keeps the array F-contiguous
            if n_indices == ary.ndim:
                # If all indices are there, ellipsis's place is indifferent
                right_indices = left_indices + right_indices
                left_indices = []
            if left_indices:
                layout = 'A'
            elif not check_contiguity(right_indices[::-1]):
                layout = 'A'

        if ndim == 0:
            # Implicitly convert to a scalar if the output ndim==0
            res = ary.dtype
        else:
            res = ary.copy(ndim=ndim, layout=layout)

    # Re-wrap indices
    if isinstance(idx, types.BaseTuple):
        idx = types.BaseTuple.from_types(all_indices)
    else:
        idx, = all_indices

    return Indexing(idx, res, advanced)


@infer_global(operator.getitem)
class GetItemBuffer(AbstractTemplate):
    def generic(self, args, kws):
        assert not kws
        [ary, idx] = args
        out = get_array_index_type(ary, idx)
        if out is not None:
            return signature(out.result, ary, out.index)


@infer_global(operator.setitem)
class SetItemBuffer(AbstractTemplate):
    def generic(self, args, kws):
        assert not kws
        ary, idx, val = args
        if not isinstance(ary, types.Buffer):
            return
        if not ary.mutable:
            msg = f"Cannot modify readonly array of type: {ary}"
            raise NumbaTypeError(msg)
        out = get_array_index_type(ary, idx)
        if out is None:
            return

        idx = out.index
        res = out.result  # res is the result type of the access ary[idx]
        if isinstance(res, types.Array):
            # Indexing produces an array
            if isinstance(val, types.Array):
                if not self.context.can_convert(val.dtype, res.dtype):
                    # DType conversion not possible
                    return
                else:
                    res = val
            elif isinstance(val, types.Sequence):
                if (res.ndim == 1 and
                    self.context.can_convert(val.dtype, res.dtype)):
                    # Allow assignment of sequence to 1d array
                    res = val
                else:
                    # NOTE: sequence-to-array broadcasting is unsupported
                    return
            else:
                # Allow scalar broadcasting
                if self.context.can_convert(val, res.dtype):
                    res = res.dtype
                else:
                    # Incompatible scalar type
                    return
        elif not isinstance(val, types.Array):
            # Single item assignment
            if not self.context.can_convert(val, res):
                # if the array dtype is not yet defined
                if not res.is_precise():
                    # set the array type to use the dtype of value (RHS)
                    newary = ary.copy(dtype=val)
                    return signature(types.none, newary, idx, res)
                else:
                    return
            res = val
        elif (isinstance(val, types.Array) and val.ndim == 0
              and self.context.can_convert(val.dtype, res)):
            # val is an array(T, 0d, O), where T is the type of res, O is order
            res = val
        else:
            return
        return signature(types.none, ary, idx, res)


def normalize_shape(shape):
    if isinstance(shape, types.UniTuple):
        if isinstance(shape.dtype, types.Integer):
            dimtype = types.intp if shape.dtype.signed else types.uintp
            return types.UniTuple(dimtype, len(shape))

    elif isinstance(shape, types.Tuple) and shape.count == 0:
        # Force (0 x intp) for consistency with other shapes
        return types.UniTuple(types.intp, 0)


@infer_getattr
class ArrayAttribute(AttributeTemplate):
    key = types.Array

    def resolve_dtype(self, ary):
        return types.DType(ary.dtype)

    def resolve_nbytes(self, ary):
        return types.intp

    def resolve_itemsize(self, ary):
        return types.intp

    def resolve_shape(self, ary):
        return types.UniTuple(types.intp, ary.ndim)

    def resolve_strides(self, ary):
        return types.UniTuple(types.intp, ary.ndim)

    def resolve_ndim(self, ary):
        return types.intp

    def resolve_size(self, ary):
        return types.intp

    def resolve_flat(self, ary):
        return types.NumpyFlatType(ary)

    def resolve_ctypes(self, ary):
        return types.ArrayCTypes(ary)

    def resolve_flags(self, ary):
        return types.ArrayFlags(ary)

    def resolve_T(self, ary):
        if ary.ndim <= 1:
            retty = ary
        else:
            layout = {"C": "F", "F": "C"}.get(ary.layout, "A")
            retty = ary.copy(layout=layout)
        return retty

    def resolve_real(self, ary):
        return self._resolve_real_imag(ary, attr='real')

    def resolve_imag(self, ary):
        return self._resolve_real_imag(ary, attr='imag')

    def _resolve_real_imag(self, ary, attr):
        if ary.dtype in types.complex_domain:
            return ary.copy(dtype=ary.dtype.underlying_float, layout='A')
        elif ary.dtype in types.number_domain:
            res = ary.copy(dtype=ary.dtype)
            if attr == 'imag':
                res = res.copy(readonly=True)
            return res
        else:
            msg = "cannot access .{} of array of {}"
            raise TypingError(msg.format(attr, ary.dtype))

    @bound_function("array.transpose")
    def resolve_transpose(self, ary, args, kws):
        def sentry_shape_scalar(ty):
            if ty in types.number_domain:
                # Guard against non integer type
                if not isinstance(ty, types.Integer):
                    msg = "transpose() arg cannot be {0}".format(ty)
                    raise TypingError(msg)
                return True
            else:
                return False

        assert not kws
        if len(args) == 0:
            return signature(self.resolve_T(ary))

        if len(args) == 1:
            shape, = args

            if sentry_shape_scalar(shape):
                assert ary.ndim == 1
                return signature(ary, *args)

            if isinstance(shape, types.NoneType):
                return signature(self.resolve_T(ary))

            shape = normalize_shape(shape)
            if shape is None:
                return

            assert ary.ndim == shape.count
            return signature(self.resolve_T(ary).copy(layout="A"), shape)

        else:
            if any(not sentry_shape_scalar(a) for a in args):
                msg = "transpose({0}) is not supported".format(
                    ', '.join(args))
                raise TypingError(msg)
            assert ary.ndim == len(args)
            return signature(self.resolve_T(ary).copy(layout="A"), *args)

    @bound_function("array.copy")
    def resolve_copy(self, ary, args, kws):
        assert not args
        assert not kws
        retty = ary.copy(layout="C", readonly=False)
        return signature(retty)

    @bound_function("array.item")
    def resolve_item(self, ary, args, kws):
        assert not kws
        # We don't support explicit arguments as that's exactly equivalent
        # to regular indexing.  The no-argument form is interesting to
        # allow some degree of genericity when writing functions.
        if not args:
            return signature(ary.dtype)

    if numpy_version < (2, 0):
        @bound_function("array.itemset")
        def resolve_itemset(self, ary, args, kws):
            assert not kws
            # We don't support explicit arguments as that's exactly equivalent
            # to regular indexing.  The no-argument form is interesting to
            # allow some degree of genericity when writing functions.
            if len(args) == 1:
                return signature(types.none, ary.dtype)

    @bound_function("array.nonzero")
    def resolve_nonzero(self, ary, args, kws):
        assert not args
        assert not kws
        if ary.ndim == 0 and numpy_version >= (2, 1):
            raise NumbaValueError(
                "Calling nonzero on 0d arrays is not allowed."
                " Use np.atleast_1d(scalar).nonzero() instead."
            )
        # 0-dim arrays return one result array
        ndim = max(ary.ndim, 1)
        retty = types.UniTuple(types.Array(types.intp, 1, 'C'), ndim)
        return signature(retty)

    @bound_function("array.reshape")
    def resolve_reshape(self, ary, args, kws):
        def sentry_shape_scalar(ty):
            if ty in types.number_domain:
                # Guard against non integer type
                if not isinstance(ty, types.Integer):
                    raise TypingError("reshape() arg cannot be {0}".format(ty))
                return True
            else:
                return False

        assert not kws
        if ary.layout not in 'CF':
            # only work for contiguous array
            raise TypingError("reshape() supports contiguous array only")

        if len(args) == 1:
            # single arg
            shape, = args

            if sentry_shape_scalar(shape):
                ndim = 1
            else:
                shape = normalize_shape(shape)
                if shape is None:
                    return
                ndim = shape.count
            retty = ary.copy(ndim=ndim)
            return signature(retty, shape)

        elif len(args) == 0:
            # no arg
            raise TypingError("reshape() take at least one arg")

        else:
            # vararg case
            if any(not sentry_shape_scalar(a) for a in args):
                raise TypingError("reshape({0}) is not supported".format(
                    ', '.join(map(str, args))))

            retty = ary.copy(ndim=len(args))
            return signature(retty, *args)

    @bound_function("array.sort")
    def resolve_sort(self, ary, args, kws):
        assert not args
        assert not kws
        return signature(types.none)

    @bound_function("array.argsort")
    def resolve_argsort(self, ary, args, kws):
        assert not args
        kwargs = dict(kws)
        kind = kwargs.pop('kind', types.StringLiteral('quicksort'))
        if not isinstance(kind, types.StringLiteral):
            raise TypingError('"kind" must be a string literal')
        if kwargs:
            msg = "Unsupported keywords: {!r}"
            raise TypingError(msg.format([k for k in kwargs.keys()]))
        if ary.ndim == 1:
            def argsort_stub(kind='quicksort'):
                pass
            pysig = utils.pysignature(argsort_stub)
            sig = signature(types.Array(types.intp, 1, 'C'), kind).replace(pysig=pysig)
            return sig

    @bound_function("array.view")
    def resolve_view(self, ary, args, kws):
        from .npydecl import parse_dtype
        assert not kws
        dtype, = args
        dtype = parse_dtype(dtype)
        if dtype is None:
            return
        retty = ary.copy(dtype=dtype)
        return signature(retty, *args)

    @bound_function("array.astype")
    def resolve_astype(self, ary, args, kws):
        from .npydecl import parse_dtype
        assert not kws
        dtype, = args
        if isinstance(dtype, types.UnicodeType):
            raise RequireLiteralValue(("array.astype if dtype is a string it "
                                       "must be constant"))
        dtype = parse_dtype(dtype)
        if dtype is None:
            return
        if not self.context.can_convert(ary.dtype, dtype):
            raise TypingError("astype(%s) not supported on %s: "
                              "cannot convert from %s to %s"
                              % (dtype, ary, ary.dtype, dtype))
        layout = ary.layout if ary.layout in 'CF' else 'C'
        # reset the write bit irrespective of whether the cast type is the same
        # as the current dtype, this replicates numpy
        retty = ary.copy(dtype=dtype, layout=layout, readonly=False)
        return signature(retty, *args)

    @bound_function("array.ravel")
    def resolve_ravel(self, ary, args, kws):
        # Only support no argument version (default order='C')
        assert not kws
        assert not args
        copy_will_be_made = ary.layout != 'C'
        readonly = not (copy_will_be_made or ary.mutable)
        return signature(ary.copy(ndim=1, layout='C', readonly=readonly))

    @bound_function("array.flatten")
    def resolve_flatten(self, ary, args, kws):
        # Only support no argument version (default order='C')
        assert not kws
        assert not args
        # To ensure that Numba behaves exactly like NumPy,
        # we also clear the read-only flag when doing a "flatten"
        # Why? Two reasons:
        # Because flatten always returns a copy. (see NumPy docs for "flatten")
        # And because a copy always returns a writeable array.
        # ref: https://numpy.org/doc/stable/reference/generated/numpy.copy.html
        return signature(ary.copy(ndim=1, layout='C', readonly=False))

    def generic_resolve(self, ary, attr):
        # Resolution of other attributes, for record arrays
        if isinstance(ary.dtype, types.Record):
            if attr in ary.dtype.fields:
                attr_dtype = ary.dtype.typeof(attr)
                if isinstance(attr_dtype, types.NestedArray):
                    return ary.copy(
                        dtype=attr_dtype.dtype,
                        ndim=ary.ndim + attr_dtype.ndim,
                        layout='A'
                    )
                else:
                    return ary.copy(dtype=attr_dtype, layout='A')


@infer_getattr
class DTypeAttr(AttributeTemplate):
    key = types.DType

    def resolve_type(self, ary):
        # Wrap the numeric type in NumberClass
        return types.NumberClass(ary.dtype)

    def resolve_kind(self, ary):
        if isinstance(ary.key, types.scalars.Float):
            val = 'f'
        elif isinstance(ary.key, types.scalars.Integer):
            val = 'i'
        else:
            return None  # other types not supported yet
        return types.StringLiteral(val)


@infer
class StaticGetItemArray(AbstractTemplate):
    key = "static_getitem"

    def generic(self, args, kws):
        # Resolution of members for record and structured arrays
        ary, idx = args
        if (isinstance(ary, types.Array) and isinstance(idx, str) and
                isinstance(ary.dtype, types.Record)):
            if idx in ary.dtype.fields:
                attr_dtype = ary.dtype.typeof(idx)
                if isinstance(attr_dtype, types.NestedArray):
                    ret = ary.copy(
                        dtype=attr_dtype.dtype,
                        ndim=ary.ndim + attr_dtype.ndim,
                        layout='A'
                    )
                    return signature(ret, *args)
                else:
                    ret = ary.copy(dtype=attr_dtype, layout='A')
                    return signature(ret, *args)


@infer_getattr
class RecordAttribute(AttributeTemplate):
    key = types.Record

    def generic_resolve(self, record, attr):
        ret = record.typeof(attr)
        assert ret
        return ret


@infer
class StaticGetItemRecord(AbstractTemplate):
    key = "static_getitem"

    def generic(self, args, kws):
        # Resolution of members for records
        record, idx = args
        if isinstance(record, types.Record) and isinstance(idx, str):
            if idx not in record.fields:
                raise NumbaKeyError(f"Field '{idx}' was not found in record "
                                    "with fields "
                                    f"{tuple(record.fields.keys())}")
            ret = record.typeof(idx)
            assert ret
            return signature(ret, *args)


@infer_global(operator.getitem)
class StaticGetItemLiteralRecord(AbstractTemplate):
    def generic(self, args, kws):
        # Resolution of members for records
        record, idx = args
        if isinstance(record, types.Record):
            if isinstance(idx, types.StringLiteral):
                if idx.literal_value not in record.fields:
                    msg = (f"Field '{idx.literal_value}' was not found in "
                           f"record with fields {tuple(record.fields.keys())}")
                    raise NumbaKeyError(msg)
                ret = record.typeof(idx.literal_value)
                assert ret
                return signature(ret, *args)
            elif isinstance(idx, types.IntegerLiteral):
                if idx.literal_value >= len(record.fields):
                    msg = f"Requested index {idx.literal_value} is out of range"
                    raise NumbaIndexError(msg)
                field_names = list(record.fields)
                ret = record.typeof(field_names[idx.literal_value])
                assert ret
                return signature(ret, *args)


@infer
class StaticSetItemRecord(AbstractTemplate):
    key = "static_setitem"

    def generic(self, args, kws):
        # Resolution of members for record and structured arrays
        record, idx, value = args
        if isinstance(record, types.Record):
            if isinstance(idx, str):
                expectedty = record.typeof(idx)
                if self.context.can_convert(value, expectedty) is not None:
                    return signature(types.void, record, types.literal(idx),
                                     value)
            elif isinstance(idx, int):
                if idx >= len(record.fields):
                    msg = f"Requested index {idx} is out of range"
                    raise NumbaIndexError(msg)
                str_field = list(record.fields)[idx]
                expectedty = record.typeof(str_field)
                if self.context.can_convert(value, expectedty) is not None:
                    return signature(types.void, record, types.literal(idx),
                                     value)


@infer_global(operator.setitem)
class StaticSetItemLiteralRecord(AbstractTemplate):
    def generic(self, args, kws):
        # Resolution of members for records
        target, idx, value = args
        if isinstance(target, types.Record) and isinstance(idx, types.StringLiteral):
            if idx.literal_value not in target.fields:
                msg = (f"Field '{idx.literal_value}' was not found in record "
                       f"with fields {tuple(target.fields.keys())}")
                raise NumbaKeyError(msg)
            expectedty = target.typeof(idx.literal_value)
            if self.context.can_convert(value, expectedty) is not None:
                return signature(types.void, target, idx, value)


@infer_getattr
class ArrayCTypesAttribute(AttributeTemplate):
    key = types.ArrayCTypes

    def resolve_data(self, ctinfo):
        return types.uintp


@infer_getattr
class ArrayFlagsAttribute(AttributeTemplate):
    key = types.ArrayFlags

    def resolve_contiguous(self, ctflags):
        return types.boolean

    def resolve_c_contiguous(self, ctflags):
        return types.boolean

    def resolve_f_contiguous(self, ctflags):
        return types.boolean


@infer_getattr
class NestedArrayAttribute(ArrayAttribute):
    key = types.NestedArray


def _expand_integer(ty):
    """
    If *ty* is an integer, expand it to a machine int (like Numpy).
    """
    if isinstance(ty, types.Integer):
        if ty.signed:
            return max(types.intp, ty)
        else:
            return max(types.uintp, ty)
    elif isinstance(ty, types.Boolean):
        return types.intp
    else:
        return ty


def generic_homog(self, args, kws):
    if args:
        raise NumbaAssertionError("args not supported")
    if kws:
        raise NumbaAssertionError("kws not supported")

    return signature(self.this.dtype, recvr=self.this)


def generic_expand(self, args, kws):
    assert not args
    assert not kws
    return signature(_expand_integer(self.this.dtype), recvr=self.this)


def sum_expand(self, args, kws):
    """
    sum can be called with or without an axis parameter, and with or without
    a dtype parameter
    """
    pysig = None
    if 'axis' in kws and 'dtype' not in kws:
        def sum_stub(axis):
            pass
        pysig = utils.pysignature(sum_stub)
        # rewrite args
        args = list(args) + [kws['axis']]
    elif 'dtype' in kws and 'axis' not in kws:
        def sum_stub(dtype):
            pass
        pysig = utils.pysignature(sum_stub)
        # rewrite args
        args = list(args) + [kws['dtype']]
    elif 'dtype' in kws and 'axis' in kws:
        def sum_stub(axis, dtype):
            pass
        pysig = utils.pysignature(sum_stub)
        # rewrite args
        args = list(args) + [kws['axis'], kws['dtype']]

    args_len = len(args)
    assert args_len <= 2
    if args_len == 0:
        # No axis or dtype parameter so the return type of the summation is a scalar
        # of the type of the array.
        out = signature(_expand_integer(self.this.dtype), *args,
                        recvr=self.this)
    elif args_len == 1 and 'dtype' not in kws:
        # There is an axis parameter, either arg or kwarg
        if self.this.ndim == 1:
            # 1d reduces to a scalar
            return_type = _expand_integer(self.this.dtype)
        else:
            # the return type of this summation is  an array of dimension one
            # less than the input array.
            return_type = types.Array(dtype=_expand_integer(self.this.dtype),
                                    ndim=self.this.ndim-1, layout='C')
        out = signature(return_type, *args, recvr=self.this)

    elif args_len == 1 and 'dtype' in kws:
        # No axis parameter so the return type of the summation is a scalar
        # of the dtype parameter.
        from .npydecl import parse_dtype
        dtype, = args
        dtype = parse_dtype(dtype)
        out = signature(dtype, *args, recvr=self.this)

    elif args_len == 2:
        # There is an axis and dtype parameter, either arg or kwarg
        from .npydecl import parse_dtype
        dtype = parse_dtype(args[1])
        return_type = dtype
        if self.this.ndim != 1:
            # 1d reduces to a scalar, 2d and above reduce dim by 1
            # the return type of this summation is  an array of dimension one
            # less than the input array.
            return_type = types.Array(dtype=return_type,
                                    ndim=self.this.ndim-1, layout='C')
        out = signature(return_type, *args, recvr=self.this)
    else:
        pass
    return out.replace(pysig=pysig)


def generic_expand_cumulative(self, args, kws):
    if args:
        raise NumbaAssertionError("args unsupported")
    if kws:
        raise NumbaAssertionError("kwargs unsupported")
    assert isinstance(self.this, types.Array)
    return_type = types.Array(dtype=_expand_integer(self.this.dtype),
                              ndim=1, layout='C')
    return signature(return_type, recvr=self.this)


def generic_hetero_real(self, args, kws):
    assert not args
    assert not kws
    if isinstance(self.this.dtype, (types.Integer, types.Boolean)):
        return signature(types.float64, recvr=self.this)
    return signature(self.this.dtype, recvr=self.this)


def generic_hetero_always_real(self, args, kws):
    assert not args
    assert not kws
    if isinstance(self.this.dtype, (types.Integer, types.Boolean)):
        return signature(types.float64, recvr=self.this)
    if isinstance(self.this.dtype, types.Complex):
        return signature(self.this.dtype.underlying_float, recvr=self.this)
    return signature(self.this.dtype, recvr=self.this)


def generic_index(self, args, kws):
    assert not args
    assert not kws
    return signature(types.intp, recvr=self.this)


def install_array_method(name, generic, prefer_literal=True):
    my_attr = {"key": "array." + name, "generic": generic,
               "prefer_literal": prefer_literal}
    temp_class = type("Array_" + name, (AbstractTemplate,), my_attr)

    def array_attribute_attachment(self, ary):
        return types.BoundFunction(temp_class, ary)

    setattr(ArrayAttribute, "resolve_" + name, array_attribute_attachment)


# Functions that return a machine-width type, to avoid overflows
install_array_method("sum", sum_expand, prefer_literal=True)


@infer_global(operator.eq)
class CmpOpEqArray(AbstractTemplate):
    #key = operator.eq

    def generic(self, args, kws):
        assert not kws
        [va, vb] = args
        if isinstance(va, types.Array) and va == vb:
            return signature(va.copy(dtype=types.boolean), va, vb)



"""towncrier_checker.py

Find missing towncrier news fragments.

Usage:
  towncrier_checker.py (-h | --help)
  towncrier_checker.py --version
  towncrier_checker.py  --token=<token> --beginning=<tag> --repo=<repo>

Options:
  -h --help          Show this screen.
  --version          Show version.
  --beginning=<tag>  Where in the History to begin
  --repo=<repo>      Which repository to look at on GitHub
  --token=<token>    The GitHub token to talk to the API

Package requirements:

- pygithub
- docopt
- gitpython

Can be installed with:

    pip install pygithub docopt gitpython

"""

import re
import os.path
from pathlib import Path

from git import Repo
from docopt import docopt
from github import Github


if __name__ == "__main__":
    arguments = docopt(__doc__, version="1.0")
    beginning = arguments["--beginning"]
    target_ghrepo = arguments["--repo"]
    github_token = arguments["--token"]
    ghrepo = Github(github_token).get_repo(target_ghrepo)
    repo = Repo(".")
    all_commits = [x for x in repo.iter_commits(f"{beginning}..HEAD")]
    merge_commits = [
        x for x in all_commits if "Merge pull request" in x.message
    ]
    prmatch = re.compile("^Merge pull request #([0-9]+) from.*")
    prs = {}
    authors = set()
    for x in merge_commits:
        match = prmatch.match(x.message)
        if match:
            issue_id = match.groups()[0]
            prs[issue_id] = "%s" % (x.message.splitlines()[2])

    # Find towncrier files and remove them from the PR list
    base = Path("docs/upcoming_changes")
    for file in base.glob("*.*.rst"):
        pr = str(file.relative_to(base)).split(".", 1)[0]
        if pr in prs:
            del prs[pr]
        else:
            print(f"No matching merge commit for {pr}")

    # Find PRs that are mentioned in CHANGE_LOG already
    # (most likely cherrypicked to previous bugfix releases)
    basepath = Path(os.path.dirname(__file__)) / ".."
    oldchangelog = basepath / "CHANGE_LOG"
    newchangelogs = (basepath / "docs" / "source" / "release").glob("*.rst")
    for path in [oldchangelog, *newchangelogs]:
        print("Checking", path)
        with open(path) as fin:
            # Use regex to find all PR numbers in the file and remove them
            pr_re = re.compile(r"\#(\d+)")
            prs_in_changelog = [
                match.group(1) for match in pr_re.finditer(fin.read())
            ]
            to_remove = {pr for pr in prs if pr in prs_in_changelog}
            for pr in to_remove:
                print(f"Removing {pr} as already in CHANGELOG")
                del prs[pr]

    # Print PRs with missing notes
    if prs:
        print("PRs that are missing towncrier notes and not skipped")
        for issue_id in sorted(prs):
            pr = ghrepo.get_pull(int(issue_id))
            if any("skip_release_notes" in label.name for label in pr.labels):
                continue  # skip
            else:
                print(issue_id, pr.html_url)



#!/usr/bin/env python
"""Download artifacts from GitHub Actions workflow runs for given run IDs."""

import argparse
import json
import os
import subprocess
import sys
from pathlib import Path


def get_artifacts_for_run(run_id, repo, token):
    """
    Get list of artifacts for a workflow run.

    Args:
        run_id: Workflow run ID
        repo: Repository in format 'owner/repo'
        token: GitHub token for API authentication

    Returns:
        List of artifact dictionaries with 'name' and 'id'
    """
    print(f"Fetching artifacts for run ID {run_id}", file=sys.stderr)

    env = os.environ.copy()
    env['GH_TOKEN'] = token

    try:
        result = subprocess.run(
            [
                "gh", "api",
                "-H", "Accept: application/vnd.github+json",
                "-H", "X-GitHub-Api-Version: 2022-11-28",
                f"/repos/{repo}/actions/runs/{run_id}/artifacts",
                "--jq", '.artifacts[] | {name: .name, id: .id}'
            ],
            capture_output=True,
            text=True,
            check=True,
            env=env
        )

        artifacts = []
        for line in result.stdout.strip().split('\n'):
            if line:
                artifacts.append(json.loads(line))

        print(f"Found {len(artifacts)} artifacts for run {run_id}", file=sys.stderr)
        return artifacts

    except subprocess.CalledProcessError as e:
        print(f"ERROR: Failed to fetch artifacts: {e}", file=sys.stderr)
        print(f"stderr: {e.stderr}", file=sys.stderr)
        return []


def download_artifact(artifact_name, artifact_id, repo, token, output_dir):
    """
    Download a single artifact and extract directly to output_dir.

    Args:
        artifact_name: Name of the artifact
        artifact_id: Artifact ID
        repo: Repository in format 'owner/repo'
        token: GitHub token for API authentication
        output_dir: Directory to download artifact to

    Returns:
        True if successful, False otherwise
    """
    print(f"Downloading artifact: {artifact_name}", file=sys.stderr)

    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)

    env = os.environ.copy()
    env['GH_TOKEN'] = token

    try:
        # Use gh CLI to download artifact
        zip_path = output_path / f"{artifact_name}.zip"
        result = subprocess.run(
            [
                "gh", "api",
                "-H", "Accept: application/vnd.github+json",
                "-H", "X-GitHub-Api-Version: 2022-11-28",
                f"/repos/{repo}/actions/artifacts/{artifact_id}/zip"
            ],
            check=True,
            env=env,
            capture_output=True
        )

        # Write the binary output to file
        with open(zip_path, 'wb') as f:
            f.write(result.stdout)

        # Unzip artifact directly to output_dir (flat structure)
        subprocess.run(
            ["unzip", "-q", "-o", str(zip_path), "-d", str(output_path)],
            check=True
        )

        # Remove the zip file
        zip_path.unlink()

        print(f"Successfully downloaded: {artifact_name}", file=sys.stderr)
        return True

    except subprocess.CalledProcessError as e:
        print(f"ERROR: Failed to download {artifact_name}: {e}", file=sys.stderr)
        if e.stderr:
            print(f"stderr: {e.stderr}", file=sys.stderr)
        if e.stdout:
            print(f"stdout: {e.stdout}", file=sys.stderr)
        return False


def main():
    parser = argparse.ArgumentParser(
        description='Download artifacts from workflow runs'
    )
    parser.add_argument(
        '--run-ids',
        nargs='+',
        required=True,
        help='List of workflow run IDs to download artifacts from'
    )
    parser.add_argument(
        '--repo',
        default=os.environ.get('GITHUB_REPOSITORY', ''),
        help='Repository in format owner/repo (default: GITHUB_REPOSITORY env var)'
    )
    parser.add_argument(
        '--token',
        default=os.environ.get('GH_TOKEN', os.environ.get('GITHUB_TOKEN', '')),
        help='GitHub token for API authentication (default: GH_TOKEN or GITHUB_TOKEN env var)'
    )
    parser.add_argument(
        '--output-dir',
        required=True,
        help='Output directory for artifacts'
    )

    args = parser.parse_args()

    if not args.repo:
        parser.error("Repository is required (via --repo or GITHUB_REPOSITORY env var)")
    if not args.token:
        parser.error("GitHub token is required (via --token or GH_TOKEN/GITHUB_TOKEN env var)")

    print(f"=== Downloading artifacts from {len(args.run_ids)} workflow runs ===", file=sys.stderr)

    for run_id in args.run_ids:
        print(f"\n--- Processing run ID: {run_id} ---", file=sys.stderr)
        artifacts = get_artifacts_for_run(run_id, args.repo, args.token)

        if not artifacts:
            print(f"\nERROR: No artifacts found for run ID {run_id}", file=sys.stderr)
            sys.exit(1)

        for artifact in artifacts:
            success = download_artifact(
                artifact['name'],
                artifact['id'],
                args.repo,
                args.token,
                args.output_dir
            )

            if not success:
                print(f"\nERROR: Failed to download artifact '{artifact['name']}'", file=sys.stderr)
                sys.exit(1)

    print(f"\n All artifacts downloaded successfully", file=sys.stderr)

if __name__ == "__main__":
    main()




from subprocess import STDOUT, check_output, CalledProcessError
import argparse
import os
import sys

manual_mode = False
parser = argparse.ArgumentParser()

parser.add_argument("--pull_request_id", required=True, type=int)
parser.add_argument("--manual", action="store_true", default=False)

args = parser.parse_args()
pr_id = args.pull_request_id
manual_mode = args.manual

rst_dir = "docs/upcoming_changes"
types_of_changes = [
    "highlight",
    "np_support",
    "deprecation",
    "expired",
    "compatibility",
    "cuda",
    "new_feature",
    "improvement",
    "performance",
    "change",
    "doc",
    "infrastructure",
    "bug_fix",
]


def list_rst_filename() -> str:
    if manual_mode:
        all_files = os.listdir(rst_dir)
        rst_file = [
            rst_dir + "/" + file
            for file in all_files
            if file.startswith(str(pr_id) + ".")
        ]
    else:
        output = check_output(
            ["git", "diff", "--name-only", "origin/release0.64"],
            encoding="utf-8",
            stderr=STDOUT,
        )
        all_files = output.strip().splitlines()
        rst_dir_modified = [
            file for file in all_files if file.startswith(rst_dir)
        ]
        print("Found modified .rst files in directory",
              f"{rst_dir} from git diff: ")
        for file in rst_dir_modified:
            print(file)
        print("\n")
        rst_file = [
            file
            for file in rst_dir_modified
            if file.startswith(rst_dir + "/" + str(args.pull_request_id) + ".")
        ]
    assert len(rst_file) == 1, (
        f"No valid .rst file found in {rst_dir} for the given"
        + f" Pull Request ID: {pr_id}. Valid .rst file should "
        + " have name starting with the respective Pull Request ID. "
        + "(see numba/docs/upcoming_changes/README for more details)"
    )
    return rst_file[0]


file = list_rst_filename()
print(f"Found required file: {file}\n")
print("Checking naming convention.")

# Must be an .rst file
assert file.endswith(".rst"), "File must be a .rst file"
# Must start file name with the PR number, followed by a ".",
# followed by type of change
filename = file.split("/")[-1]

all_towncrier_rst = os.listdir(rst_dir)
all_towncrier_rst = [
    rst
    for rst in all_towncrier_rst
    if not (rst.startswith("template") or rst.startswith("README"))
]
all_pr_ids = [int(rst.split(".")[0]) for rst in all_towncrier_rst]
assert len(set(all_pr_ids)) == len(
    all_pr_ids
), "All PR IDs must be unique. Please check for duplicate PR IDs"

assert (
    len(filename.split(".")) == 3
), "Filename must be in the format <PR_ID>.<type_of_change>.rst"

# Must be one of the required types of changes
assert filename.split(".")[1] in types_of_changes, (
    "File must be one of the following types of changes:"
    + " highlight, np_support, deprecation, expired, compatibility"
    + ", cuda, new_feature, improvement, performance, change, doc"
    + ", infrastructure, bug_fix"
)
print("Passed: Filename is valid\n")

print("Checking file contents:\n")
# Check rst contents
with open(file, "r") as f:
    contents = f.read().splitlines()
    print("\n".join(contents))
    # First line must be the title followed by the underline
    assert len(contents) >= 4, "File must have at least four lines"
    title = contents[0]
    assert len(title) > 0, "Title must not be empty"
    underline = contents[1]
    for underline_type in underline:
        assert underline_type == "-", \
            "Header should be underlined with - characters"
    assert len(title) == len(underline), (
        "Title and underline must be the same length."
        + f" (Found Title: {len(title)}, Underline: {len(underline)})"
    )
    blank_line = contents[2]
    assert len(blank_line) == 0, "Third line must be blank"
    description = contents[3]
    assert len(description) > 0, "Description must not be empty"
    print("\nPassed: File contents are valid\n")

print("Validating RST")
try:
    output = check_output(["rstcheck", file], stderr=STDOUT, encoding="utf-8")
except CalledProcessError as cpe:
    print("Failed: rstcheck failed")
    print(cpe.output)
    sys.exit(cpe.returncode)
else:
    assert "Success! No issues detected." in output
    print("Passed: rstcheck passed")



"""gitlog2changelog.py

Usage:
  gitlog2changelog.py (-h | --help)
  gitlog2changelog.py --version
  gitlog2changelog.py --token=<token> --beginning=<tag> --repo=<repo>

Options:
  -h --help          Show this screen.
  --version          Show version.
  --beginning=<tag>  Where in the History to begin
  --repo=<repo>      Which repository to look at on GitHub
  --token=<token>    The GitHub token to talk to the API

Dependencies:
  * docopt
  * pygithub
  * gitpython

"""

import re

from git import Repo
from docopt import docopt
from github import Github, Auth


def hyperlink_user(user_obj):
    try:
        return "`%s <%s>`_" % (user_obj.login, user_obj.html_url)
    except AttributeError:
        return user_obj


def author_key(x):
    return x.login.lower() if not isinstance(x, str) else x.lower()


if __name__ == '__main__':

    # Process command line arguments.
    arguments = docopt(__doc__, version='1.0')
    beginning = arguments['--beginning']
    target_ghrepo = arguments['--repo']
    github_token = arguments['--token']

    # this is an object representation of the repo on github using pygithub
    ghrepo = Github(auth=Auth.Token(github_token)).get_repo(target_ghrepo)
    # this is an object representation of the local git repo using gitpython
    repo = Repo('.')

    # # We obtain the merge commits for the given range by looking at the
    # commit message. If it begins with the string "Merge pull request" it is
    # either a true merge commit (with two parents) that originats from merging
    # a pull-requets OR it is a cherry-picked merge-commit. Both are fine and
    # should be included.
    merge_commits = [x for x in repo.iter_commits(f'{beginning}..HEAD')
                   if x.message.startswith('Merge pull request')]
    # We use the following regular expression to extract the GitHub ID of the
    # pull-request from the message.
    prmatch = re.compile(
        f'^Merge pull request #(\\d+) from.*')

    # Then, using the list, we use the regular expression above to extract the
    # ID of the pull-requests and build a dictionary mapping the Pull-Request
    # ID to it's description.
    pull_requests = {}
    for m in merge_commits:
        match = prmatch.match(m.message)
        number = int(match.group(1))
        # There are very rare cases where the commit message of the merge
        # commit created when merging a pull-request on GitHub does not include
        # the description. In that case, we use the GitHub API to obtain that
        # description.
        try:
            description = m.message.splitlines()[2]
        except IndexError:
            p = ghrepo.get_pull(number)
            description = p.title
        # Lastly, add the pull-request to the dictionary.
        pull_requests[number] = str(description)

    # Now that we have the pull-requests, we iterate through them to determine
    # the authors involved.
    authors = set()
    print("Pull-Requests:\n")
    for k in sorted(pull_requests.keys()):
        pr = ghrepo.get_pull(k)
        hyperlink = "`#%s <%s>`_" % (k, pr.html_url)
        # get all users for all commits
        pr_authors = set()
        for c in pr.get_commits():
            # Here we need to try to get some kind of an "author" string for
            # the given commit. The first thing we try is to get the GitHub
            # username associated with the email address in the commit
            # metadata. If this fails, when someone has used a 'user.email'
            # setting that does not correspond to a known account on GitHub, we
            # attempt to get the `user.name` setting from the commit directly.
            # We need to convert the commit from its pygithub style object to
            # its gitpython style object in order to access that information.
            # We do the same for the committer.
            author = (c.author if c.author is not None
                      else repo.commit(str(c.sha)).author.name)
            assert author is not None and author != ""
            pr_authors.add(author)
            committer = (c.committer
                         if (c.committer is not None
                             and c.committer.login != "web-flow")
                         else repo.commit(str(c.sha)).committer.name)
            if committer != "GitHub":
                pr_authors.add(committer)
        print("* PR %s: %s (%s)" % (hyperlink, pull_requests[k],
                                    " ".join([hyperlink_user(u) for u in
                                              sorted(pr_authors, key=author_key)])))
        for a in pr_authors:
            authors.add(a)

    print("")
    print("Authors:\n")
    [print('* %s' % hyperlink_user(x)) for x in sorted(authors, key=author_key)]



#!/usr/bin/env python
"""Find successful llvmlite workflow run IDs for a given git tag."""

import argparse
import json
import os
import subprocess
import sys


def find_workflow_run(workflow_name, tag, repo, token):
    """
    Find the latest successful workflow run for a given workflow and tag.

    Args:
        workflow_name: Name of the workflow file (e.g., 'llvmlite_conda_builder.yml')
        tag: Git tag to search for
        repo: Repository in format 'owner/repo' (e.g., 'numba/llvmlite')
        token: GitHub token for API authentication

    Returns:
        Run ID as a string, or None if not found
    """
    print(f"Search for successful run of {workflow_name} on tag {tag} in {repo}", file=sys.stderr)

    env = os.environ.copy()
    env['GH_TOKEN'] = token

    try:
        result = subprocess.run(
            [
                "gh", "api",
                "-H", "Accept: application/vnd.github+json",
                "-H", "X-GitHub-Api-Version: 2022-11-28",
                f"/repos/{repo}/actions/workflows/{workflow_name}/runs",
                "--jq", f'.workflow_runs[] | select(.head_branch == "{tag}" and .conclusion == "success") | .id'
            ],
            capture_output=True,
            text=True,
            check=True,
            env=env
        )

        run_ids = result.stdout.strip().split('\n')
        if run_ids and run_ids[0]:
            run_id = run_ids[0]
            print(f"Found run ID {run_id} for {workflow_name}", file=sys.stderr)
            return run_id

        print(f"ERROR: No successful run found for {workflow_name} on tag {tag}", file=sys.stderr)
        return None

    except subprocess.CalledProcessError as e:
        print(f"ERROR: Failed to query GitHub API: {e}", file=sys.stderr)
        print(f"stderr: {e.stderr}", file=sys.stderr)
        return None


def load_workflow_config(config_path):
    """
    Load workflow groups from JSON config.

    Args:
        config_path: Path to JSON file containing workflow groups under 'workflows' key

    Returns:
        Dict mapping group names to lists of workflow filenames
    """
    try:
        with open(config_path, 'r') as f:
            config = json.load(f)
    except FileNotFoundError:
        print(f"ERROR: Config file not found: {config_path}", file=sys.stderr)
        sys.exit(1)
    except json.JSONDecodeError as e:
        print(f"ERROR: Invalid JSON in config file: {e}", file=sys.stderr)
        sys.exit(1)

    try:
        workflow_groups = config['workflows']
    except KeyError as e:
        print(f"ERROR: Missing required key in config file: {e}", file=sys.stderr)
        sys.exit(1)

    # Validate that all groups are lists
    for group_name, workflows in workflow_groups.items():
        if not isinstance(workflows, list):
            print(f"ERROR: Workflow group '{group_name}' must be a list", file=sys.stderr)
            sys.exit(1)

    return workflow_groups


def main():
    parser = argparse.ArgumentParser(
        description='Find successful workflow run IDs for a given tag'
    )
    parser.add_argument(
        '--tag',
        default=os.environ.get('TAG', ''),
        help='Git tag to search for (default: TAG env var)'
    )
    parser.add_argument(
        '--repo',
        default=os.environ.get('GITHUB_REPOSITORY', ''),
        help='Repository in format owner/repo (default: GITHUB_REPOSITORY env var)'
    )
    parser.add_argument(
        '--token',
        default=os.environ.get('GH_TOKEN', os.environ.get('GITHUB_TOKEN', '')),
        help='GitHub token for API authentication (default: GH_TOKEN or GITHUB_TOKEN env var)'
    )
    parser.add_argument(
        '--config',
        required=True,
        help='Path to JSON config file containing workflow groups'
    )

    args = parser.parse_args()

    if not args.tag:
        parser.error("TAG is required (via --tag or TAG env var)")
    if not args.repo:
        parser.error("Repository is required (via --repo or GITHUB_REPOSITORY env var)")
    if not args.token:
        parser.error("GitHub token is required (via --token or GH_TOKEN/GITHUB_TOKEN env var)")

    # Load workflow groups from config file
    # Returns dict: {group_name: [workflow_file1.yml, workflow_file2.yml, ...]}
    grouped_workflow_names = load_workflow_config(args.config)

    # Will store found run IDs: {group_name: [run_id1, run_id2, ...]}
    found_run_ids_by_group = {}

    # For each workflow group, query GitHub API to find successful run IDs
    # matching the specified tag
    for group_name, workflow_files in grouped_workflow_names.items():
        print(f"\n=== finding {group_name} workflow runs ===", file=sys.stderr)
        run_ids = []
        for workflow_file in workflow_files:
            run_id = find_workflow_run(workflow_file, args.tag, args.repo, args.token)
            if run_id:
                run_ids.append(run_id)
            else:
                print(f"WARNING: Could not find run ID for {workflow_file}", file=sys.stderr)
        found_run_ids_by_group[group_name] = run_ids

    # Verify all workflows were found
    total_expected = sum(len(workflows) for workflows in grouped_workflow_names.values())
    total_found = sum(len(run_ids) for run_ids in found_run_ids_by_group.values())

    if total_found != total_expected:
        print(f"\nERROR: Not all workflow runs were found ({total_found}/{total_expected})", file=sys.stderr)
        sys.exit(1)

    print("\n=== found workflow run IDs ===", file=sys.stderr)
    for group_name, run_ids in found_run_ids_by_group.items():
        print(f"{group_name}: {' '.join(run_ids)}", file=sys.stderr)

    # Output run IDs in shell variable format for GitHub Actions
    # eg. wheel_run_ids=run_id1 run_id2 ...
    for group_name, run_ids in found_run_ids_by_group.items():
        print(f"{group_name}_run_ids={' '.join(run_ids)}")


if __name__ == "__main__":
    main()




import sys
import re
import os
import subprocess as subp


known_rpaths = {
    '@rpath/libc++.1.dylib': "/usr/lib/libc++.1.dylib",
}


def run_shell(cmd):
    out = subp.check_output(cmd, shell=True)
    return out


def main(whl):
    # Find shared libraries
    run_shell('wheel unpack {}'.format(whl))
    thedir = '-'.join(whl.split('-')[:2])

    sharedlibs = run_shell('find {} -name "*.so"'.format(thedir)).decode().splitlines()
    # Scan paths
    regex_rpath = re.compile(r'@rpath\/[^ ]+')
    updated_files = set()
    for path in sharedlibs:
        # Check rpaths
        otool_out = run_shell('otool -L {}'.format(path)).decode()
        rpaths = set(regex_rpath.findall(otool_out))
        unknown = rpaths - set(known_rpaths.keys())
        if unknown:
            #raise AssertionError('unknown rpath {}'.format(unknown))
            print('unknown', unknown)
        # Patch
        for rpath in rpaths & set(known_rpaths.keys()):
            print('patch', rpath, 'in', path)
            run_shell('install_name_tool -change {} {} {}'.format(
                rpath, known_rpaths[rpath], path))
            updated_files.add(path)
    print('updated_files', updated_files)
    # Update whl
    if updated_files:
        run_shell('wheel pack {}'.format(thedir))
    else:
        print('nothing to do')


if __name__ == '__main__':
    main(*sys.argv[1:])
